
from ...squeal.tables.pipeline import Pipeline


def build_pipeline(log, session, type, factory, date, schedule=None):
    for cls, args, kargs in Pipeline.all(log, session, type):
        log.info('Building %s (%s, %s)' % (cls, args, kargs))
        yield from cls(log).build(session, factory, date, *args, schedule=schedule, **kargs)


class Displayer:

    def __init__(self, log):
        self._log = log





from tempfile import NamedTemporaryFile

from sqlalchemy.sql.functions import count

from ch2.command.activities import activities
from ch2.command.args import bootstrap_file, m, V, DEV, mm, FAST
from ch2.command.constants import constants
from ch2.config.default import default
from ch2.squeal.tables.activity import ActivityJournal
from ch2.squeal.tables.pipeline import PipelineType
from ch2.squeal.tables.statistic import StatisticJournal
from ch2.stoats.calculate import run_pipeline_after


def test_activities():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        args, log, db = bootstrap_file(f, m(V), '5', 'constants', '--set', 'FTHR.%', '154')
        constants(args, log, db)

        args, log, db = bootstrap_file(f, m(V), '5', 'constants', 'FTHR.%')
        constants(args, log, db)

        args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                       'activities', mm(FAST), 'data/test/personal/2018-08-27-rec.fit')
        activities(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:
            n = s.query(count(StatisticJournal.id)).scalar()
            assert n == 27, n
            journal = s.query(ActivityJournal).one()
            assert journal.time != journal.finish





from collections import defaultdict

from pandas import DataFrame
from pygeotile.point import Point
from sqlalchemy.sql.functions import count

from ..squeal.database import connect
from ..squeal.tables.activity import ActivityGroup, ActivityJournal
from ..squeal.tables.monitor import MonitorJournal, MonitorHeartRate, MonitorSteps
from ..squeal.tables.source import Interval, Source
from ..squeal.tables.statistic import StatisticName, StatisticJournal, StatisticMeasure


def extract(data, instance, *attributes):
    for attribute in attributes:
        if hasattr(instance, attribute):
            data[attribute].append(getattr(instance, attribute))
        else:
            data[attribute].append(None)


class Data:
    '''
    Provide access to the database via DataFrames.
    Intended for use in Jupyter notebooks.
    All data retrieved in a single (read-only) session.
    '''

    def __init__(self, log, db):
        self._log = log
        self._session = db.session()

    def activities(self):
        data, ids = defaultdict(list), []
        for activity_group in self._session.query(ActivityGroup).order_by(ActivityGroup.name):
            ids.append(activity_group.id)
            extract(data, activity_group, 'name', 'description')
            data['count'].append(self._session.query(count(ActivityJournal.id)).
                                 filter(ActivityJournal.activity_group == activity_group).scalar())
        return DataFrame(data, index=ids)

    def activity_journals(self, activity, start=None, finish=None):
        data, times = defaultdict(list), []
        q = self._session.query(ActivityJournal).join(ActivityGroup).filter(ActivityGroup.name == activity)
        if start:
            q = q.filter(ActivityJournal.time >= start)
        if finish:
            q = q.filter(ActivityJournal.time <= finish)
        for journal in q.order_by(ActivityJournal.time).all():
            times.append(journal.time)
            extract(data, journal, 'name', 'fit_file', 'finish')
        return DataFrame(data, index=times)

    def activity_waypoints(self, activity, time):
        journal = self._session.query(ActivityJournal).join(ActivityGroup). \
            filter(ActivityGroup.name == activity, ActivityJournal.time == time).one()
        frames = []
        for timespan in journal.timespans:
            data, times = defaultdict(list), []
            for waypoint in timespan.waypoints:
                times.append(waypoint.time)
                extract(data, waypoint, 'latitude', 'longitude', 'heart_rate', 'distance', 'speed')
                if waypoint.latitude and waypoint.longitude:
                    p = Point.from_latitude_longitude(waypoint.latitude, waypoint.longitude)
                    x, y = p.meters
                else:
                    x, y = None, None
                data['x'].append(x)
                data['y'].append(y)
            frames.append(DataFrame(data, index=times))
        return frames

    def statistics(self, *statistics):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        data = defaultdict(list)
        for statistic in self._session.query(StatisticName). \
                filter(StatisticName.id.in_(statistic_ids)).order_by(StatisticName.name):
            extract(data, statistic, 'name', 'description', 'units', 'summary', 'owner', 'constraint')
            data['count'].append(self._session.query(count(StatisticJournal.id)).
                                 filter(StatisticJournal.statistic_name == statistic).scalar())
        return DataFrame(data)

    def _collect_statistics(self, statistics):
        if not statistics:
            statistics = ['%']
        statistic_ids, statistic_names = set(), set()
        for statistic_name in statistics:
            for statistic_name in self._session.query(StatisticName). \
                    filter(StatisticName.name.like(statistic_name)).all():
                statistic_ids.add(statistic_name.id)
                statistic_names.add(statistic_name.name)
        return statistic_names, statistic_ids

    def _build_statistic_journal_query(self, statistic_ids, q,
            start, finish, owner, constraint):
        q = q.filter(StatisticName.id.in_(statistic_ids))
        if start:
            q = q.filter(Source.time >= start)
        if finish:
            q = q.filter(Source.time <= finish)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        if constraint:
            q = q.filter(StatisticName.constraint == constraint)
        return q

    def statistic_journals(self, *statistics, start=None, finish=None, owner=None, constraint=None, schedule=None):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        q = self._build_statistic_journal_query(
            statistic_ids, self._session.query(StatisticJournal).join(StatisticName, Source),
            start, finish, owner, constraint)
        if schedule:
            q = q.join(Interval).filter(Interval.schedule == schedule)
        raw_data = defaultdict(dict)
        for journal in q.all():  # todo - eager load
            raw_data[journal.time][journal.statistic_name.name] = journal.value
        data, times = defaultdict(list), []
        for time in sorted(raw_data.keys()):
            times.append(time)
            sub_data = raw_data[time]
            for statistic in statistic_names:
                if statistic in sub_data:
                    data[statistic].append(sub_data[statistic])
                else:
                    data[statistic].append(None)
        return DataFrame(data, index=times)

    def statistic_quartiles(self, *statistics, schedule='m', start=None, finish=None, owner=None, constraint=None):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        q = self._session.query(StatisticMeasure). \
            join(StatisticJournal, StatisticMeasure.statistic_journal_id == StatisticJournal.id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            join(Source, StatisticJournal.source_id == Source.id)
        q = self._build_statistic_journal_query(statistic_ids, q, start, finish, owner, constraint)
        if schedule:
            q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
                filter(Interval.schedule == schedule)
        q = q.filter(StatisticMeasure.quartile != None)
        raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
        for measure in q.all():
            raw_data[measure.source.time][measure.statistic_journal.statistic_name.name][measure.quartile] = \
                measure.statistic_journal.value
        data, times = defaultdict(list), []
        for time in sorted(raw_data.keys()):
            times.append(time)
            sub_data = raw_data[time]
            for statistic in statistic_names:
                if statistic in sub_data:
                    data[statistic].append(sub_data[statistic])
                else:
                    data[statistic].append(None)
        return DataFrame(data, index=times)

    def monitor_journals(self):
        data, times = defaultdict(list), []
        for journal in self._session.query(MonitorJournal).order_by(MonitorJournal.time).all():
            times.append(journal.time)
            extract(data, journal, 'fit_file', 'finish')
        return DataFrame(data, index=times)

    def monitor_steps(self, time):
        data, times = defaultdict(list), []
        for steps in self._session.query(MonitorSteps).join(MonitorJournal). \
                filter(MonitorJournal.time == time). \
                order_by(MonitorSteps.time).all():
            times.append(steps.time)
            extract(data, steps, 'value', 'activity')
        return DataFrame(data, index=times)

    def monitor_heart_rate(self, time):
        data, times = defaultdict(list), []
        for heart_rate in self._session.query(MonitorHeartRate).join(MonitorJournal). \
                filter(MonitorJournal.time == time). \
                order_by(MonitorHeartRate.time).all():
            times.append(heart_rate.time)
            extract(data, heart_rate, 'value')
        return DataFrame(data, index=times)


def data(*args):
    '''
    Start here to access data.  Create an instance in Jupyter:

        from ch2.data import data
        d = data('-v', '4')
        d.statistics()
        ...
    '''
    ns, log, db = connect(args)
    return Data(log, db)




from .database import Counter, add_statistics, add_activity_group, add_activity_constant, add_topic, add_topic_field, \
    add_diary, add_activities, add_monitor
from ..lib.schedule import Schedule
from ..squeal.tables.statistic import StatisticJournalType
from ..stoats.calculate.activity import ActivityStatistics
from ..stoats.calculate.clean import CleanUnusedStatistics
from ..stoats.calculate.monitor import MonitorStatistics
from ..stoats.calculate.summary import SummaryStatistics
from ..stoats.display.activity import ActivityDiary
from ..stoats.names import BPM, FTHR
from ..stoats.read.activity import ActivityImporter
from ..stoats.read.monitor import MonitorImporter
from ..uweird.fields.topic import Text, Float, Score0


def default(db):

    with db.session_context() as s:

        # the following users heleper functions (add_...) but you can also
        # execute arbitrary python code, use the session, etc.

        # statistics pipeline (called to calculate missing statistics)

        c = Counter()
        add_statistics(s, ActivityStatistics, c)
        add_statistics(s, MonitorStatistics, c)
        # need to call normalize here because schedule isn't a schedule type column,
        # but part of a kargs JSON blob.
        add_statistics(s, SummaryStatistics, c, schedule=Schedule.normalize('m'))
        add_statistics(s, SummaryStatistics, c, schedule=Schedule.normalize('y'))
        add_statistics(s, CleanUnusedStatistics, c)

        # diary pipeline (called to display data in the diary)

        c = Counter()
        add_diary(s, ActivityDiary, c)

        # monitor pipeline

        add_monitor(s, MonitorImporter, c)

        # basic activities

        bike = add_activity_group(s, 'Bike', c, description='All cycling activities')
        run = add_activity_group(s, 'Run', c, description='All running activities')
        # sport_to_activity maps from the FIT sport field to the activity defined above
        add_activities(s, ActivityImporter, c, sport_to_activity={'cycling': bike.name,
                                                                  'running': run.name})
        s.flush()  # set IDs because these are used below

        # constants used by statistics

        add_activity_constant(s, bike, FTHR,
                              description='Heart rate at functional threshold (cycling). See https://www.britishcycling.org.uk/knowledge/article/izn20140808-Understanding-Intensity-2--Heart-Rate-0',
                              units=BPM, type=StatisticJournalType.INTEGER)
        add_activity_constant(s, run, FTHR,
                              description='Heart rate at functional threshold (running).',
                              units=BPM, type=StatisticJournalType.INTEGER)

        # a basic diary

        c = Counter()
        diary = add_topic(s, 'Diary', c)
        add_topic_field(s, diary, 'Notes', c,
                        display_cls=Text)
        # now provided via monitor
        # add_topic_field(s, diary, 'Rest HR', c,
        #                 units=BPM, summary='[avg]',
        #                 display_cls=Integer, lo=25, hi=75)
        add_topic_field(s, diary, 'Weight', c,
                        units='kg', summary='[avg]',
                        display_cls=Float, lo=50, hi=100, dp=1)
        add_topic_field(s, diary, 'Sleep', c,
                        units='h', summary='[avg]',
                        display_cls=Float, lo=0, hi=24, dp=1)
        add_topic_field(s, diary, 'Mood', c,
                        summary='[avg]',
                        display_cls=Score0)
        add_topic_field(s, diary, 'Nutrition', c,
                        summary='[cnt]',
                        display_cls=Text)
        add_topic_field(s, diary, 'Soreness', c,
                        summary='[cnt]',
                        display_cls=Text)
        add_topic_field(s, diary, 'Medication', c,
                        summary='[cnt]',
                        display_cls=Text)
        add_topic_field(s, diary, 'Weather', c,
                        summary='[cnt]',
                        display_cls=Text)
        add_topic_field(s, diary, 'Route', c,
                        summary='[cnt]',
                        display_cls=Text)




from _csv import reader
from glob import glob
from itertools import zip_longest
from logging import getLogger, basicConfig, DEBUG
from os.path import splitext, split, join
from sys import stdout
from tempfile import TemporaryDirectory

from ch2.command.args import FIELDS
from ch2.fit.format.read import filtered_records
from ch2.fit.format.records import no_names, append_units, no_bad_values, fix_degrees, chain
from ch2.fit.profile.fields import DynamicField
from ch2.fit.profile.profile import read_profile
from ch2.fit.summary import summarize, summarize_csv, summarize_tables


def test_profile():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()
    nlog, types, messages = read_profile(log, '/home/andrew/project/ch2/choochoo/data/sdk/Profile.xlsx')

    cen = types.profile_to_type('carry_exercise_name')
    assert cen.profile_to_internal('farmers_walk') == 1

    session = messages.profile_to_message('session')
    field = session.profile_to_field('total_cycles')
    assert isinstance(field, DynamicField), type(field)
    for name in field.references:
        assert name == 'sport'

    workout_step = messages.profile_to_message('workout_step')
    field = workout_step.number_to_field(4)
    assert field.name == 'target_value', field.name
    fields = ','.join(sorted(field.references))
    assert fields == 'duration_type,target_type', fields


def test_decode():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()
    data, types, messages, records = \
        filtered_records(log, '/home/andrew/project/ch2/choochoo/data/test/personal/2018-07-26-rec.fit',
                         profile_path='/home/andrew/project/ch2/choochoo/data/sdk/Profile.xlsx')
    for record in records:
        print(record.into(tuple, filter=chain(no_names, append_units, no_bad_values, fix_degrees)))


def test_dump():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()
    summarize(log, FIELDS,
              '/home/andrew/project/ch2/choochoo/data/test/personal/2018-07-30-rec.fit',
              profile_path='/home/andrew/project/ch2/choochoo/data/sdk/Profile.xlsx')


def test_developer():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()
    summarize(log, FIELDS,
              '/home/andrew/project/ch2/choochoo/data/test/sdk/DeveloperData.fit',
              profile_path='/home/andrew/project/ch2/choochoo/data/sdk/Profile.xlsx')


def dump_csv(log, fit_file, csv_file):
    with open(csv_file, 'w') as output:
        summarize_csv(log, fit_file,
                      profile_path='/home/andrew/project/ch2/choochoo/data/sdk/Profile.xlsx',
                      out=output)


# https://docs.python.org/3/library/itertools.html#itertools-recipes
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


class Skip:

    def __init__(self, skip):
        self.skip = skip

    def __bool__(self):
        try:
            return bool(self.skip)
        finally:
            self.skip -= 1


def cleaned(values):
    # want a good example for scaling errors, not zero
    # also, suspect some 0.0 might be 0.... (remove this later to check)
    return (value[:-2] if value.endswith('.0') else value for value in values)


def compare_rows(log, us, them, name, skip):
    assert us[0:3] == them[0:3] or skip, "%s != %s for %s\n(%s\n%s)" % (us[0:3], them[0:3], name, us, them)
    excess = len(them) % 3
    if excess and not any(them[-excess:]):
        log.warn('Discarding %d empty values from reference' % excess)
        them = them[:-excess]
    while len(them) > len(us) + 2 and not any(them[-3:]):
        them = them[:-3]
    # after first 3 entries need to sort to be sure order is correct
    for us_nvu, them_nvu in zip_longest(sorted(grouper(cleaned(us[3:]), 3)),
                                        sorted(grouper(cleaned(them[3:]), 3))):
        if 'COMPOSITE' not in us_nvu[1]:
            assert us_nvu == them_nvu or skip, "%s != %s for %s\n(%s\n%s)" % (us_nvu, them_nvu, name, us, them)


def compare_csv(log, us, them, name, skip):
    # print(us)
    # with open(us, 'r') as us_in:
    #     for line in us_in.readlines():
    #         print("%s" % line.strip())
    #         if '\0' in line:
    #             print(sub('\0', 'NULL', line.strip()))
    # print(them)
    # with open(them, 'r') as them_in:
    #     for line in them_in.readlines():
    #         print("%s" % line.strip())
    #         if '\0' in line:
    #             print(sub('\0', 'NULL', line.strip()))
    with open(us, 'r') as us_in, open(them, 'r') as them_in:
        us_reader = reader(us_in)
        them_reader = reader(them_in)
        next(them_reader)  # skip titles
        for us_row, them_row in zip_longest(us_reader, them_reader):
            compare_rows(log, us_row, them_row, name, skip)


def test_csv():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()

    with TemporaryDirectory() as dir:
        skip = Skip(2)  # timer_trigger in ActivityGroup.fit
        for fit_file in glob('/home/andrew/project/ch2/choochoo/data/test/sdk/*.fit'):
            print(fit_file)
            fit_dir, file = split(fit_file)
            name = splitext(file)[0]
            csv_name = '%s.%s' % (name, 'csv')
            csv_us = join(dir, csv_name)
            csv_them = join(fit_dir, csv_name)
            dump_csv(log, fit_file, csv_us)
            compare_csv(log, csv_us, csv_them, name, skip)
        assert skip.skip == 0, 'Unused skipped tests'


def test_personal():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()

    with TemporaryDirectory() as dir:
        skip = Skip(2)  # timer_trigger in ActivityGroup.fit
        for fit_file in glob('/home/andrew/project/ch2/choochoo/data/test/personal/*.fit'):
            print(fit_file)
            summarize_tables(log, fit_file)


def test_timestamp_16():

    basicConfig(stream=stdout, level=DEBUG)
    log = getLogger()
    data, types, messages, records = \
        filtered_records(log, '/home/andrew/project/ch2/choochoo/data/test/personal/andrew@acooke.org_24755630065.fit',
                         profile_path='/home/andrew/project/ch2/choochoo/data/sdk/Profile.xlsx')
    for record in records:
        if record.name == 'monitoring':
            print(record.into(tuple, filter=chain(no_names, append_units, no_bad_values, fix_degrees)))





from tempfile import NamedTemporaryFile

import matplotlib.pyplot as plt

from ch2.command.activities import activities
from ch2.command.constants import constants
from ch2.command.args import bootstrap_file, m, mm, DEV, V, F
from ch2.config.default import default
from ch2.data import data, col_to_boxstats


def import_activity(f):
    bootstrap_file(f, m(V), '0', mm(DEV), configurator=default)
    args, log, db = bootstrap_file(f, m(V), '0', 'constants', '--set', 'FTHR.%', '154')
    constants(args, log, db)
    args, log, db = bootstrap_file(f, m(V), '0', mm(DEV),
                                   'activities', 'data/test/personal/2018-08-27-rec.fit')
    activities(args, log, db)


def test_activity():
    with NamedTemporaryFile() as f:
        import_activity(f)
        # there's a bokeh version of this in the notebooks
        d = data(m(V), '0', mm(DEV), m(F), f.name)
        activities = d.activities()
        print(activities)
        statistics = d.statistics()
        print(statistics)
        journals = d.statistic_journals('Active %')
        print(journals)
        # run('sqlite3 %s ".dump"' % f.name, shell=True)


# below assumes data in the local DB

# def test_activities():
#     d = data()
#     frame = d.statistic_journals('Active Distance')
#     print(frame)
#     fig, ax = plt.subplots()
#     plt.scatter(x=frame.index, y=frame['Active Distance'] / 1000)
#     plt.xlabel('Date')
#     plt.ylabel('Distance / km')
#     plt.title('Active Distance')
#     fig.autofmt_xdate()
#     plt.savefig('/tmp/distance.png')
#
#
# def test_quartiles():
#     # there's a bokeh version of this in the notebooks
#     d = data('-v', '0')
#     stats = d.statistics()
#     print(stats)
#     frame = d.statistic_quartiles('Max Med HR over 30m')
#     print(frame)
#     stats = col_to_boxstats(frame, 'Max Med HR over 30m')
#     print(stats)
#     # https://matplotlib.org/gallery/statistics/bxp.html
#     fig, ax = plt.subplots()
#     ax.bxp(stats, showfliers=False)
#     fig.autofmt_xdate()
#     plt.savefig('/tmp/summary.png')
#
#
# def test_route():
#     d = data('-v', '0')
#     frames = d.activity_waypoints('Bike', '2018-09-06')
#     print(frames)
#     # can't find a good, portable solution past this point
#     # within jupyter, bokeh works ok, but you can't save images without
#     # getting into problems with node (javascript).  seems to be a general
#     # problem of python map libraries.




from abc import ABC, abstractmethod

from urwid import Text

from ...lib.utils import label, em

PAGE_WIDTH = 4


class Field(ABC):

    def __init__(self, log, journal, width=1):
        self._log = log
        self._journal = journal
        self.width = width

    def __str__(self):
        if self._journal.value is None:
            return '%s: _' % self._journal.statistic_name.name + self._format_units()
        else:
            return '%s: ' % self._journal.statistic_name.name + \
                   self._format_value(self._journal.value) + self._format_units()

    @abstractmethod
    def _format_value(self, value):
        pass

    def _format_units(self):
        return (' ' + self._journal.statistic_name.units) if self._journal.statistic_name.units else ''

    def widget(self):
        return self._widget()

    @abstractmethod
    def _widget(self):
        pass


class ReadOnlyField(Field):

    def __init__(self, log, journal, width=1, date=None):
        self._date = date
        super().__init__(log, journal, width=width)

    def _format_value(self, value):
        return self._journal.formatted()

    def _widget(self):
        if self._date:
            measures = self._journal.measures_as_text(self._date)
        else:
            measures = []
        if measures:
            self.width += 1
        return Text([label('%s: ' % self._journal.statistic_name.name), em(self._journal.formatted()), ' ']
                    + measures)







from .command.activities import activities
from .command.args import COMMAND, parser, NamespaceWithVariables, PROGNAME, HELP, DEV, DIARY, FIT, \
    PACKAGE_FIT_PROFILE, ACTIVITIES, NO_OP, DEFAULT_CONFIG, CONSTANTS, STATISTICS, TEST_SCHEDULE, MONITOR, GARMIN
from .command.constants import constants
from .command.diary import diary
from .command.default_config import default_config
from .command.fit import fit
from .command.garmin import garmin
from .command.help import help, LengthFmt
from .command.monitor import monitor
from .command.package_fit_profile import package_fit_profile
from .command.statistics import statistics
from .command.test_schedule import test_schedule
from .lib.log import make_log
from .squeal.database import Database


def no_op(args, log, db):
    '''
## no-op

This is used internally when accessing data in Jupyter or configuring the system
at the command line.
    '''
    pass


COMMANDS = {ACTIVITIES: activities,
            CONSTANTS: constants,
            DEFAULT_CONFIG: default_config,
            DIARY: diary,
            FIT: fit,
            GARMIN: garmin,
            HELP: help,
            MONITOR: monitor,
            STATISTICS: statistics,
            NO_OP: no_op,
            PACKAGE_FIT_PROFILE: package_fit_profile,
            TEST_SCHEDULE: test_schedule,
            }


def main():
    p = parser()
    args = NamespaceWithVariables(p.parse_args())
    command_name = args[COMMAND] if COMMAND in args else None
    command = COMMANDS[command_name] if command_name in COMMANDS else None
    tui = command and hasattr(command, 'tui') and command.tui
    log = make_log(args, tui=tui)
    db = Database(args, log)
    try:
        if db.is_empty() and (not command or command_name != DEFAULT_CONFIG):
            request_config()
        elif command:
            command(args, log, db)
        else:
            log.debug('If you are seeing the "No command given" error during development ' +
                      'you may have forgotten to set the command name via `set_defaults()`.')
            raise Exception('No command given (try `ch2 help`)')
    except KeyboardInterrupt:
        log.critical('User abort')
        pass
    except Exception as e:
        log.critical(e)
        log.info('See `%s %s` for available commands.' % (PROGNAME, HELP))
        log.info('Docs at http://andrewcooke.github.io/choochoo')
        if not args or args[DEV]:
            raise


def request_config():
    LengthFmt().print_all('''
Welcome to Choochoo.

Before using the ch2 command you must configure the system.

Please see the documentation at http://andrewcooke.github.io/choochoo

To generate a default configuration use the command

    %s %s

NOTE: The default configuration is only an example.  Please see the docs
for more details.''' % (PROGNAME, DEFAULT_CONFIG))




from ..config import default


def default_config(args, log, db):
    '''
## default-config

    > ch2 default-config

Generate a simple initial configuration.

Please see the documentation at http://andrewcooke.github.io/choochoo
    '''
    default(db)




ACTIVE_DISTANCE = 'Active Distance'
ACTIVE_TIME = 'Active Time'
ACTIVE_SPEED = 'Active Speed'
FTHR = 'FTHR'
MEDIAN_KM_TIME = 'Median %dkm Time'
MEDIAN_KM_TIME_ANY = 'Median % Time'
PERCENT_IN_Z = 'Percent in Z%d'
PERCENT_IN_Z_ANY = 'Percent in Z%'
TIME_IN_Z = 'Time in Z%d'
MAX_MED_HR_M = 'Max Med HR %dm'
MAX_MED_HR_M_ANY = 'Max Med HR %'
HR_MINUTES = (5, 10, 15, 20, 30, 60, 90, 120, 180)

MAX = '[max]'
MIN = '[min]'
CNT = '[cnt]'
AVG = '[avg]'
SUM = '[sum]'

M = 'm'
S = 's'
H = 'h'
KMH = 'km/h'
PC = '%'
BPM = 'bpm'
KG = 'kg'


def round_km():
    yield from range(5, 21, 5)
    yield from range(25, 76, 25)
    yield from range(100, 251, 50)
    yield from range(300, 1001, 100)


STEPS = 'Steps'
REST_HR = 'Rest HR'



import datetime as dt
from re import search

from urwid import Text, Pile, Columns, Divider

from . import Displayer
from .heart_rate import build_zones
from ..calculate.activity import ActivityStatistics
from ..names import ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED, MEDIAN_KM_TIME_ANY, MAX_MED_HR_M_ANY
from ...lib.date import format_seconds, local_date_to_time
from ...lib.utils import label
from ...squeal.tables.activity import ActivityGroup, ActivityJournal
from ...squeal.tables.source import Source
from ...squeal.tables.statistic import StatisticJournal, StatisticName
from ch2.uweird.fields.summary import summary_columns
from ...uweird.tui.decorators import Indent

HRZ_WIDTH = 30


class ActivityDiary(Displayer):

    def build(self, s, f, date, schedule=None):
        if schedule:
            yield from self.__build_schedule(s, f, date, schedule)
        else:
            yield from self.__build_date(s, f, date)

    def __build_date(self, s, f, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for journal in s.query(ActivityJournal). \
                    filter(ActivityJournal.time >= start,
                           ActivityJournal.time < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.time).all():
                yield self.__journal_date(s, journal, date)

    def __journal_date(self, s, ajournal, date):
        zones = build_zones(s, ajournal, HRZ_WIDTH)
        details = Pile(self.__active_date(s, ajournal, date))
        return Pile([
            Text(ajournal.name),
            Indent(Columns([details, (HRZ_WIDTH + 2, zones)])),
            Divider(),
            Indent(Columns([Pile(self.__template(s, ajournal, MEDIAN_KM_TIME_ANY,
                                                 'Median Time', r'(\d+km)', date)),
                            Pile(self.__template(s, ajournal, MAX_MED_HR_M_ANY,
                                                 'Max Med HR', r'(\d+m)', date))]))
        ])

    def __active_date(self, s, ajournal, date):
        body = [Divider()]
        body.append(Text('%s - %s  (%s)' % (ajournal.time.strftime('%H:%M:%S'), ajournal.finish.strftime('%H:%M:%S'),
                                            format_seconds((ajournal.finish - ajournal.time).seconds))))
        for name in (ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED):
            sjournal = StatisticJournal.at_time(s, ajournal.time, name, ActivityStatistics, ajournal.activity_group.id)
            body.append(Text([label('%s: ' % sjournal.statistic_name.name)] + self.__format_value(sjournal, date)))
        return body

    def __template(self, s, ajournal, template, title, re, date):
        body = [Text(title)]
        sjournals = s.query(StatisticJournal).join(StatisticName, Source). \
            filter(Source.time == ajournal.time,
                   StatisticName.name.like(template),
                   StatisticName.owner == ActivityStatistics,
                   StatisticName.constraint == ajournal.activity_group.id).order_by(StatisticName.name).all()
        # extract
        for sjournal in self.__sort_journals(sjournals):
            body.append(Text([label(search(re, sjournal.statistic_name.name).group(1) + ': ')] +
                             self.__format_value(sjournal, date)))
        return body

    def __sort_journals(self, sjournals):
        return sorted(sjournals,
                      # order by integer embedded in name
                      key=lambda sjournal: int(search(r'(\d+)', sjournal.statistic_name.name).group(1)))

    def __sort_names(self, statistic_names):
        return sorted(statistic_names,
                      # order by integer embedded in name
                      key=lambda statistic_name: int(search(r'(\d+)', statistic_name.name).group(1)))

    def __format_value(self, sjournal, date):
        return ['%s ' % sjournal.formatted()] + sjournal.measures_as_text(date)

    def __build_schedule(self, s, f, date, schedule):
        columns = list(self.__schedule_fields(s, f, date, schedule))
        if columns:
            yield Pile([Text('Activities'),
                        Indent(Pile(columns))])

    def __schedule_fields(self, s, f, date, schedule):
        names = list(self.__names(s, ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED))
        yield from summary_columns(self._log, s, f, date, schedule, names)
        names = self.__sort_names(self.__names_like(s, MEDIAN_KM_TIME_ANY))
        yield from summary_columns(self._log, s, f, date, schedule, names)
        names = self.__sort_names(self.__names_like(s, MAX_MED_HR_M_ANY))
        yield from summary_columns(self._log, s, f, date, schedule, names)

    def __names(self, s, *names):
        for name in names:
            yield s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == ActivityStatistics).one()

    def __names_like(self, s, name):
        return s.query(StatisticName). \
                filter(StatisticName.name.like(name),
                       StatisticName.owner == ActivityStatistics).all()




from argparse import ArgumentParser
from genericpath import exists
from os import makedirs
from os.path import dirname, expanduser, realpath, normpath, relpath, join
from re import compile, sub
from typing import Mapping

from ch2.lib.date import to_date

PROGNAME = 'ch2'
COMMAND = 'command'
TOPIC = 'topic'

ACTIVITIES = 'activities'
CONSTANTS = 'constants'
DEFAULT_CONFIG = 'default-config'
DIARY = 'diary'
FIT = 'fit'
GARMIN = 'garmin'
H, HELP = 'h', 'help'
NO_OP = 'no-op'
PACKAGE_FIT_PROFILE = 'package-fit-profile'
STATISTICS = 'statistics'
TEST_SCHEDULE = 'test-schedule'

ACTIVITY = 'activity'
AFTER = 'after'
ALL_MESSAGES = 'all-messages'
ALL_FIELDS = 'all-fields'
CSV = 'csv'
DATABASE = 'database'
DATE = 'date'
DELETE = 'delete'
DEV = 'dev'
DIR = 'dir'
F = 'f'
FAST = 'fast'
FIELDS = 'fields'
FINISH = 'finish'
FORMAT = 'format'
FTHR = 'fthr'
FORCE, F = 'force', 'f'
GREP = 'grep'
LIKE = 'like'
LIMIT = 'limit'
LOGS = 'logs'
LIST = 'list'
M, MESSAGE = 'm', 'message'
MATCH = 'match'
MESSAGES = 'messages'
MONITOR = 'monitor'
MONTH = 'month'
MONTHS = 'months'
NAME = 'name'
NOT = 'not'
PASS = 'pass'
PATH = 'path'
PLAN = 'plan'
PWD = 'pwd'
RECORDS = 'records'
ROOT = 'root'
SERVICE = 'service'
SET = 'set'
SCHEDULE = 'schedule'
START = 'start'
TABLES = 'tables'
USER = 'user'
V, VERBOSITY = 'v', 'verbosity'
VALUE = 'value'
VERSION = 'version'
W, WARN = 'w', 'warn'
YEAR = 'year'


def mm(name): return '--' + name
def m(name): return '-' + name


VARIABLE = compile(r'(.*(?:[^$]|^))\${(\w+)\}(.*)')
MEMORY = ':memory:'


class NamespaceWithVariables(Mapping):

    def __init__(self, ns):
        self._dict = vars(ns)

    def __getitem__(self, name):
        name = sub('-', '_', name)
        value = self._dict[name]
        try:
            match = VARIABLE.match(value)
            while match:
                value = match.group(1) + self[match.group(2)] + match.group(3)
                match = VARIABLE.match(value)
            return sub(r'\$\$', '$', value)
        except TypeError:
            return value

    def path(self, name, index=None, rooted=True):
        # special case sqlite3 in-memory database
        if self[name] == MEMORY: return self[name]
        path = self[name]
        if index is not None: path = path[index]
        path = expanduser(path)
        if rooted and relpath(path) and name != ROOT:
            path = join(self.path(ROOT), path)
        return realpath(normpath(path))

    def file(self, name, index=None, rooted=True):
        file = self.path(name, index=index, rooted=rooted)
        # special case sqlite3 in-memory database
        if file == MEMORY: return file
        path = dirname(file)
        if not exists(path):
            makedirs(path)
        return file

    def dir(self, name, index=None, rooted=True):
        path = self.path(name, index=index, rooted=rooted)
        if not exists(path):
            makedirs(path)
        return path

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self.__dict__)


def parser():

    parser = ArgumentParser(prog=PROGNAME)

    parser.add_argument(m(F), mm(DATABASE), action='store', default='${root}/database.sqlj', metavar='FILE',
                        help='the database file')
    parser.add_argument(mm(DEV), action='store_true', help='enable development mode')
    parser.add_argument(mm(LOGS), action='store', default='logs', metavar='DIR',
                        help='the directory for logs')
    parser.add_argument(mm(ROOT), action='store', default='~/.ch2', metavar='DIR',
                        help='the root directory for the default configuration')
    parser.add_argument(m(V), mm(VERBOSITY), action='store', nargs=1, default=None, type=int, metavar='VERBOSITY',
                        help='output level for stderr (0: silent; 5:noisy)')
    parser.add_argument(mm(VERSION), action='version', version='0.4.3',
                        help='display version and exit')

    subparsers = parser.add_subparsers()

    activities = subparsers.add_parser(ACTIVITIES, help='read activity data')
    activities.add_argument(mm(FORCE), action='store_true', help='re-read file and delete existing data')
    activities.add_argument(mm(FAST), action='store_true', help='do not calculate statistics')
    activities.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                            help='path to fit file(s)')
    activities.set_defaults(command=ACTIVITIES)

    constant = subparsers.add_parser(CONSTANTS, help='set and examine constants')
    constant_flags = constant.add_mutually_exclusive_group()
    constant_flags.add_argument(mm(DELETE), action='store_true', help='delete existing value(s)')
    constant_flags.add_argument(mm(SET), action='store_true', help='store a new value')
    constant.add_argument(mm(FORCE), action='store_true', help='confirm deletion(s) without value')
    constant.add_argument(NAME, action='store', nargs='?', metavar='NAME', help='constant name')
    constant.add_argument(DATE, action='store', nargs='?', metavar='DATE', help='date when measured')
    constant.add_argument(VALUE, action='store', nargs='?', metavar='VALUE', help='constant value')
    constant.set_defaults(command=CONSTANTS)

    basic_config = subparsers.add_parser(DEFAULT_CONFIG,
                                         help='configure the default database ' +
                                              '(see docs for full configuration instructions)')
    basic_config.set_defaults(command=DEFAULT_CONFIG)

    diary = subparsers.add_parser(DIARY, help='daily diary and summary')
    diary.add_argument(DATE, action='store', metavar='DATE', nargs='?', type=to_date,
                       help='an optional date to display (default is today)')
    diary_summary = diary.add_mutually_exclusive_group()
    diary_summary.add_argument(mm(MONTH), action='store_const', dest=SCHEDULE, const='m',
                               help='show monthly summary')
    diary_summary.add_argument(mm(YEAR), action='store_const', dest=SCHEDULE, const='y',
                               help='show yearly summary')
    diary_summary.add_argument(mm(SCHEDULE), metavar='SCHEDULE',
                               help='show summary for given schedule')
    diary.set_defaults(command=DIARY)

    fit = subparsers.add_parser(FIT, help='display contents of fit file')
    fit.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                     help='path to fit file')
    fit_format = fit.add_mutually_exclusive_group(required=True)
    fit_format.add_argument(mm(RECORDS), action='store_const', dest=FORMAT, const=RECORDS,
                            help='show high-level structure (ordered by time)')
    fit_format.add_argument(mm(TABLES), action='store_const', dest=FORMAT, const=TABLES,
                            help='show high-level structure (grouped in tables)')
    fit_format.add_argument(mm(GREP), action='store', dest=GREP, nargs='+', metavar='MSG:FLD[=VAL]',
                            help='show med-level matching entries')
    fit_format.add_argument(mm(CSV), action='store_const', dest=FORMAT, const=CSV,
                            help='show med-level structure in CSV format')
    fit_format.add_argument(mm(MESSAGES), action='store_const', dest=FORMAT, const=MESSAGES,
                            help='show low-level message structure')
    fit_format.add_argument(mm(FIELDS), action='store_const', dest=FORMAT, const=FIELDS,
                            help='show low-level field structure (more details)')
    fit.add_argument(mm(AFTER), action='store', type=int, metavar='N', default=0,
                     help='skip initial messages')
    fit.add_argument(mm(LIMIT), action='store', type=int, metavar='N', default=-1,
                     help='limit number of messages')
    fit.add_argument(mm(ALL_FIELDS), action='store_true',
                     help='display undocumented high-level fields')
    fit.add_argument(mm(ALL_MESSAGES), action='store_true',
                     help='display undocumented high-level messages')
    fit.add_argument(m(M), mm(MESSAGE), action='store', nargs='+', metavar='MSG',
                     help='display only named high-level messages')
    fit.add_argument(m(W), mm(WARN), action='store_true',
                     help='additional warning messages')
    fit.add_argument(mm(NAME), action='store_true',
                     help='display file name')
    fit.add_argument(mm(NOT), action='store_true',
                     help='display file names that don\'t match (with --grep --name)')
    fit.add_argument(mm(MATCH), action='store', type=int, default=1,
                     help='number of matches to display (with --grep, default 1, -1 for all)')
    fit.set_defaults(command=FIT, format=GREP)   # because that's the only one not set if the option is used

    garmin = subparsers.add_parser(GARMIN, help='download monitor data from garmin connect')
    garmin.add_argument(DIR, action='store', metavar='DIR',
                        help='the directory where FIT files are stored')
    garmin.add_argument(mm(USER), action='store', metavar='USER', required=True,
                        help='garmin connect username')
    garmin.add_argument(mm(PASS), action='store', metavar='PASSWORD', required=True,
                        help='garmin connect password')
    garmin.add_argument(mm(DATE), action='store', metavar='DATE', type=to_date,
                        help='date to download')
    garmin.set_defaults(command=GARMIN)

    help = subparsers.add_parser(HELP, help='display help')
    help.add_argument(TOPIC, action='store', nargs='?', metavar=TOPIC,
                      help='the subject for help')
    help.set_defaults(command=HELP)

    monitor = subparsers.add_parser(MONITOR, help='read monitor data')
    monitor.add_argument(mm(FORCE), action='store_true', help='re-read file and delete existing data')
    monitor.add_argument(mm(FAST), action='store_true', help='do not calculate statistics')
    monitor.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                         help='path to fit file(s)')
    monitor.set_defaults(command=MONITOR)

    statistics = subparsers.add_parser(STATISTICS, help='(re-)generate statistics')
    statistics.add_argument(mm(FORCE), action='store_true', help='delete existing statistics')
    statistics.add_argument(AFTER, action='store', nargs='?', metavar='DATE',
                            help='date from which statistics are deleted')
    statistics.add_argument(mm(LIKE), action='store', metavar='PATTERN',
                            help='run only matching pipeline classes')
    statistics.set_defaults(command=STATISTICS)

    noop = subparsers.add_parser(NO_OP,
                                 help='used within jupyter (no-op from cmd line)')
    noop.set_defaults(command=NO_OP)

    package_fit_profile = subparsers.add_parser(PACKAGE_FIT_PROFILE,
                                                help='parse and save the global fit profile (dev only)')
    package_fit_profile.add_argument(PATH, action='store', metavar='PROFILE',
                                     help='the path to the profile (Profile.xlsx)')
    package_fit_profile.add_argument(m(W), mm(WARN), action='store_true',
                                     help='additional warning messages')
    package_fit_profile.set_defaults(command=PACKAGE_FIT_PROFILE)

    test_schedule = subparsers.add_parser(TEST_SCHEDULE, help='print schedule locations in a calendar')
    test_schedule.add_argument(SCHEDULE, action='store', metavar='SCHEDULE',
                               help='the schedule to test')
    test_schedule.add_argument(mm(START), action='store', metavar='DATE',
                               help='the date to start displaying data')
    test_schedule.add_argument(mm(MONTHS), action='store', metavar='N', type=int,
                               help='the number of months to display')
    test_schedule.set_defaults(command=TEST_SCHEDULE)

    return parser


def bootstrap_file(file, *args, configurator=None, post_config=None):

    from ..lib.log import make_log
    from ..config.database import config
    from ..squeal.database import Database

    args = [mm(DATABASE), file.name] + list(args)
    if configurator:
        log, db = config(*args)
        configurator(db)
    args += post_config if post_config else []
    args = NamespaceWithVariables(parser().parse_args(args))
    log = make_log(args)
    db = Database(args, log)

    return args, log, db




from sqlalchemy import Column, Text

from ch2.squeal.support import Base
from ch2.squeal.types import Time, Owner


class FileScan(Base):

    __tablename__ = 'file_scan'

    path = Column(Text, nullable=False, primary_key=True)
    owner = Column(Owner, nullable=False, primary_key=True)
    md5_hash = Column(Text, nullable=False, index=True)
    last_scan = Column(Time, nullable=False)




from logging import getLogger, basicConfig, DEBUG, StreamHandler, Formatter
from sys import stdout

from sqlalchemy import Column, Integer, ForeignKey, create_engine
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.sql.functions import count

from ch2.squeal.support import Base

basicConfig()
log = getLogger()
log.setLevel(DEBUG)
handler = StreamHandler(stdout)
handler.setLevel(DEBUG)
formatter = Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
handler.setFormatter(formatter)
log.addHandler(handler)


class Database:

    def __init__(self):
        self._log = log
        self.engine = create_engine('sqlite:///:memory:', echo=True)
        listen(self.engine, 'connect', self.fk_pragma_on_connect)
        self.session = sessionmaker(bind=self.engine)
        Base.metadata.create_all(self.engine)

    @staticmethod
    def fk_pragma_on_connect(dbapi_con, con_record):
        dbapi_con.execute('pragma foreign_keys=ON')


class Parent(Base):

    __tablename__ = 'parent'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)
    parent_ref_id = Column(Integer, ForeignKey('other.id', ondelete='cascade'))
    parent_ref = relationship('Other')

    __mapper_args__ = {
        'polymorphic_identity': 0,
        'polymorphic_on': type
    }


class Child(Parent):

    __tablename__ = 'child'

    id = Column(Integer, ForeignKey('parent.id', ondelete='cascade'), primary_key=True)
    child_ref_id = Column(Integer, ForeignKey('other.id', ondelete='cascade'))
    child_ref = relationship('Other')

    __mapper_args__ = {
        'polymorphic_identity': 1,
        'passive_deletes': True
    }


class Other(Base):

    __tablename__ = 'other'

    id = Column(Integer, primary_key=True)


def setup():

    db = Database()
    s = db.session()
    o1 = Other()
    s.add(o1)
    o2 = Other()
    s.add(o2)
    s.add(Child(parent_ref=o1, child_ref=o2))
    s.commit()
    return db


def assert_eq(a, b):
    assert a == b, '%s != %s' % (a, b)


def test_delete_child_instance():

    db = setup()
    s = db.session()
    c = s.query(Child).one()
    s.delete(c)
    assert_eq(s.query(count(Parent.id)).scalar(), 0)


def test_delete_child_sql():

    db = setup()
    s = db.session()
    s.query(Child).delete()
    # !!! this is because the ondelete cascade goes "the other way"
    assert_eq(s.query(count(Parent.id)).scalar(), 1)


def test_delete_parent_instance():

    db = setup()
    s = db.session()
    p = s.query(Parent).one()
    s.delete(p)
    assert_eq(s.query(count(Child.id)).scalar(), 0)


def test_delete_parent_sql():

    db = setup()
    s = db.session()
    s.query(Parent).delete()
    assert_eq(s.query(count(Child.id)).scalar(), 0)


def test_delete_child_ref_instance():

    db = setup()
    s = db.session()
    o = s.query(Child).one().child_ref
    s.delete(o)
    s.commit()
    s = db.session()
    # !!! this is because SQL deletes the child via cascade, but not the parent (see test_delete_child_sql)
    # (we don't have a backref on Other, so SQLAlchemy isn't going to do anything - have no idea if it would)
    assert_eq(s.query(count(Parent.id)).scalar(), 1)
    assert_eq(s.query(count(Child.id)).scalar(), 0)


def test_delete_parent_ref_instance():

    db = setup()
    s = db.session()
    o = s.query(Child).one().parent_ref
    s.delete(o)
    s.commit()
    s = db.session()
    assert_eq(s.query(count(Parent.id)).scalar(), 0)
    assert_eq(s.query(count(Child.id)).scalar(), 0)




from abc import abstractmethod

from ch2.lib.io import for_modified_files


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class Importer:

    def __init__(self, log, db):
        self._log = log
        self._db = db

    def _first(self, path, records, *names):
        try:
            return next(iter(record for record in records if record.name in names))
        except StopIteration:
            self._log.warn('No %s entry(s) in %s' % (str(names), path))
            raise AbortImportButMarkScanned()

    def _last(self, path, records, *names):
        save = None
        for record in records:
            if record.name in names:
                save = record
        if not save:
            self._log.warn('No %s entry(s) in %s' % (str(names), path))
            raise AbortImportButMarkScanned()
        return save

    def _run(self, paths, force=False, **kargs):
        with self._db.session_context() as s:
            for_modified_files(self._log, s, paths, self._callback(kargs), self, force=force)

    def _callback(self, kargs):
        def callback(file):
            self._log.info('Scanning %s' % file)
            with self._db.session_context() as s:
                try:
                    self._import(s, file, **kargs)
                    return True
                except AbortImport as e:
                    self._log.debug('Aborted %s' % file)
                    return isinstance(e, AbortImportButMarkScanned)
        return callback

    @abstractmethod
    def _import(self, s, path, **kargs):
        pass




from os.path import splitext, basename

from ..read import AbortImport
from ...fit.format.read import filtered_records
from ...fit.format.records import fix_degrees
from ...lib.date import to_time
from ...lib.utils import datetime_to_epoch
from ...squeal.database import add
from ...squeal.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan, ActivityWaypoint
from ...stoats.read import Importer


class ActivityImporter(Importer):

    def run(self, paths, force=False, sport_to_activity=None):
        if sport_to_activity is None:
            raise Exception('No map from sport to activity')
        self._run(paths, force=force, sport_to_activity=sport_to_activity)

    def _activity_group(self, s, path, sport, sport_to_activity):
        if sport in sport_to_activity:
            return self._lookup_activity_group(s, sport_to_activity[sport])
        else:
            self._log.warn('Unrecognised sport: "%s" in %s' % (sport, path))
            raise AbortImport()

    def _lookup_activity_group(self, s, name):
        activity_group = s.query(ActivityGroup).filter(ActivityGroup.name == name).one_or_none()
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                self._log.info('Available activities:')
                for activity_group in activities:
                    self._log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                self._log.error('No activities defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _delete_journals(self, s, activity_group, first_timestamp):
        # need to iterate because sqlite doesn't support multi-table delete and we have inheritance.
        for journal in s.query(ActivityJournal). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.time == first_timestamp).all():
            s.delete(journal)
        s.flush()

    def _import(self, s, path, sport_to_activity):

        data, types, messages, records = filtered_records(self._log, path)
        records = [record.force(fix_degrees)
                   for record in sorted(records, key=lambda r: r.timestamp if r.timestamp else to_time(0.0))]

        first_timestamp = self._first(path, records, 'event', 'record').value.timestamp
        sport = self._first(path, records, 'sport').value.sport.lower()
        activity_group = self._activity_group(s, path, sport, sport_to_activity)
        self._delete_journals(s, activity_group, first_timestamp)
        journal = add(s, ActivityJournal(activity_group=activity_group, time=first_timestamp,
                                         fit_file=path, name=splitext(basename(path))[0]))

        timespan, warned, latest = None, 0, to_time(0.0)
        for record in records:
            try:
                if record.name == 'event' or (record.name == 'record' and record.timestamp > latest):
                    if record.name == 'event' and record.value.event == 'timer' and record.value.event_type == 'start':
                        timespan = add(s, ActivityTimespan(activity_journal=journal,
                                                           start=datetime_to_epoch(record.value.timestamp)))
                    if record.name == 'record':
                        waypoint = add(s, ActivityWaypoint(activity_journal=journal,
                                                           activity_timespan=timespan,
                                                           time=datetime_to_epoch(record.value.timestamp),
                                                           latitude=record.none.position_lat,
                                                           longitude=record.none.position_long,
                                                           heart_rate=record.none.heart_rate,
                                                           # if distance is not set in some future file, calculate from
                                                           # lat/long?
                                                           distance=record.value.distance,
                                                           speed=record.none.enhanced_speed))
                    if record.name == 'event' and record.value.event == 'timer' \
                            and record.value.event_type == 'stop_all':
                        if timespan:
                            timespan.finish = datetime_to_epoch(record.value.timestamp)
                            journal.finish = record.value.timestamp
                            timespan = None
                        else:
                            self._log.warn('Ignoring stop with no corresponding start (possible lost data?)')
                    if record.name == 'record':
                        latest = record.timestamp
                else:
                    if record.name == 'record':
                        self._log.warn('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                       (path, record.timestamp))
            except (AttributeError, TypeError) as e:
                if warned < 10:
                    self._log.warn('Error while reading %s - some data may be missing (%s)' % (path, e))
                elif warned == 10:
                    self._log.warn('No more warnings will be given for %s' % path)
                warned += 1



from tempfile import NamedTemporaryFile

from ch2.command.args import bootstrap_file, m, V, mm, DEV
from ch2.config.default import default
from ch2.config.plan.british import twelve_week_improver
from ch2.config.plan.exponential import exponential_time, exponential_distance
from ch2.lib.date import to_date, add_date
from ch2.squeal.tables.topic import Topic


def test_british():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        plan = twelve_week_improver('2018-07-25')
        plan.create(log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:
            root = s.query(Topic).filter(Topic.parent_id == None, Topic.name == 'Plan').one()
            assert len(root.children) == 1, root.children
            assert root.schedule
            assert root.schedule.start == to_date('2018-07-25'), root.schedule.start
            assert root.schedule.finish == add_date('2018-07-25', (12, 'w')), root.schedule.finish
            parent = root.children[0]
            assert len(parent.children) == 7, parent.children
            for child in parent.children:
                print(child)


def test_exponential_time():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        plan = exponential_time('Time test', '2d[2]', '20M', 5, '2018-07-25', '3m')
        plan.create(log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:
            root = s.query(Topic).filter(Topic.parent_id == None, Topic.name == 'Plan').one()
            assert len(root.children) == 1, root.children
            parent = root.children[0]
            assert len(parent.children) == 46, len(parent.children)
            for child in parent.children:
                print(child)


def test_exponential_distance():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        plan = exponential_distance('Distance test', 'w[mon,wed,fri]', '20km', 5, '2018-07-25', '1m')
        plan.create(log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:
            root = s.query(Topic).filter(Topic.parent_id == None, Topic.name == 'Plan').one()
            assert len(root.children) == 1, root.children
            parent = root.children[0]
            assert len(parent.children) == 14, len(parent.children)
            for child in parent.children:
                print(child)




from ...squeal.tables.pipeline import Pipeline


def run_pipeline_after(log, db, type, after=None, force=False, like=None):
    with db.session_context() as s:
        for cls, args, kargs in Pipeline.all(log, s, type, like=like):
            log.info('Running %s (%s, %s)' % (cls, args, kargs))
            cls(log, db).run(*args, force=force, after=after, **kargs)


def run_pipeline_paths(log, db, type, paths, force=False, like=None):
    with db.session_context() as s:
        for cls, args, kargs in Pipeline.all(log, s, type, like=like):
            log.info('Running %s (%s, %s)' % (cls, args, kargs))
            cls(log, db).run(paths, *args, force=force, **kargs)




from tempfile import NamedTemporaryFile

import sqlalchemy.sql.functions as func

from ch2 import monitor
from ch2.command.args import bootstrap_file, m, V, DEV, mm, FAST
from ch2.config.default import default
from ch2.lib.date import to_time, to_date, local_date_to_time
from ch2.squeal.tables.monitor import MonitorJournal, MonitorSteps
from ch2.squeal.tables.pipeline import PipelineType
from ch2.squeal.tables.source import Source, SourceType, Interval
from ch2.squeal.tables.statistic import StatisticJournal, StatisticName
from ch2.stoats.calculate import run_pipeline_after
from ch2.stoats.calculate.monitor import MonitorStatistics
from ch2.stoats.names import STEPS, REST_HR


def test_monitor():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                       'monitor', mm(FAST), 'data/test/personal/25822184777.fit')
        monitor(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:
            n = s.query(func.count(StatisticJournal.id)).scalar()
            assert n == 10, n
            mjournal = s.query(MonitorJournal).one()
            assert mjournal.time != mjournal.finish


def test_values():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        for file in ('24696157869', '24696160481', '24696163486'):
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                           'monitor', mm(FAST), 'data/test/personal/andrew@acooke.org_%s.fit' % file)
            monitor(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:

            mjournals = s.query(MonitorJournal).order_by(MonitorJournal.time).all()
            assert mjournals[2].time == to_time('2018-09-06 15:06:00'), mjournals[2].time
            print(mjournals[2].fit_file)

            # steps

            steps = mjournals[2].steps
            assert len(steps) == 72, len(steps)
            assert steps[0].value == 1336, steps[0].value
            assert steps[1].value == 355, steps[1].value
            total = sum(s.delta for s in steps)
            assert total == 11066, total
            summary = s.query(StatisticJournal).join(StatisticName, Source, Interval). \
                filter(Source.type == SourceType.INTERVAL,
                       Source.time == local_date_to_time(to_date('2018-09-06')),
                       Interval.schedule == 'd',
                       StatisticName.owner == MonitorStatistics,
                       StatisticName.name == STEPS).one()
            # connect has 12757 for this date,
            assert summary.value == 12757, summary.value
            check = s.query(func.sum(MonitorSteps.delta)). \
                filter(MonitorSteps.time >= to_time('2018-09-06 03:00'),
                       MonitorSteps.time < to_time('2018-09-07 03:00')).scalar()
            assert check == 12757, check

            # heart rate

            hrs = sorted(mjournals[2].heart_rate, key=lambda x: x.time)
            assert len(hrs) == 571, len(hrs)
            assert hrs[0].value == 74, hrs[0].value
            hrs = [hr.value for hr in hrs]
            assert min(*hrs) == 0, min(*hrs)
            hrs = [hr for hr in hrs if hr > 0]
            assert min(*hrs) == 52, min(*hrs)
            summary = s.query(StatisticJournal).join(StatisticName, Source, Interval). \
                filter(Source.type == SourceType.INTERVAL,
                       Source.time == to_time('2018-09-06 03:00'),
                       Interval.schedule == 'd',
                       StatisticName.owner == MonitorStatistics,
                       StatisticName.name == REST_HR).one()
            assert summary.value == 42, summary.value


def test_bug():

     with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        for file in ('25505915679', '25519562859', '25519565531', '25532154264', '25539076032', '25542112328'):
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                           'monitor', mm(FAST), 'data/test/personal/andrew@acooke.org_%s.fit' % file)
            monitor(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:

            # steps

            summary = s.query(StatisticJournal).join(StatisticName, Source, Interval). \
                filter(Source.type == SourceType.INTERVAL,
                       Source.time == local_date_to_time(to_date('2018-10-07')),
                       Interval.schedule == 'd',
                       StatisticName.owner == MonitorStatistics,
                       StatisticName.name == STEPS).one()

            # connect has 3031 for this date.
            assert summary.value == 3031, summary.value




from abc import abstractmethod
from sys import stdout

from .args import TOPIC, HELP, PROGNAME, m, H, DEFAULT_CONFIG
from ..lib.io import terminal_width


class Fmt:

    def print_all(self, text):
        self.print()
        for para in self.paras(text):
            self.print(para)
        self.print()

    def paras(self, text):
        para = ''
        for line in text.split('\n'):
            if line.strip().startswith('*'):
                if para:
                    yield para
                para = line
            elif para and not line:
                yield para
                yield ''
                para = ''
            elif para:
                para += ' ' + line
            else:
                para = line
        if para:
            yield para

    @abstractmethod
    def print(self, para):
        raise NotImplementedError()


class LengthFmt(Fmt):

    def __init__(self, stream=stdout, width=None, margin=1):
        self.__out = stream
        self.__width = terminal_width(width) - margin
        self.__margin = margin

    def print(self, text=None):
        if text:
            indent = ' ' * self.__margin
            while text and text[0] == ' ':
                indent += ' '
                text = text[1:]
            line = indent
            while text:
                try:
                    word, text = text.split(' ', 1)
                except ValueError:
                    word, text = text, ''
                if len(line) == len(indent):
                    line += word
                elif len(line) + 1 + len(word) <= self.__width:
                    line += ' ' + word
                else:
                    print(line, file=self.__out)
                    line = indent + word
            print(line, file=self.__out)
        else:
            print(file=self.__out)


def commands(COMMANDS):
    return '''
Thank-you for using Choochoo.  Please send feedback to andrew@acooke.org

# Commands

* %s

See also `%s %s` for usage, '%s %s CMD` for guidance on a particular command, 
and `%s %s CMD` for usage of that command.

Docs at http://andrewcooke.github.io/choochoo/index''' % (
        '\n* '.join(COMMANDS.keys()), PROGNAME, m(H), PROGNAME, HELP, PROGNAME, m(H))


def help(args, logs, db):
    '''
## help

    > ch2 help [topic]

Displays help for a particular topic.

### Examples

    > ch2 help help

Displays this information.

    > ch2 help

Lists available topics.
    '''
    from .. import COMMANDS
    if args[TOPIC] in COMMANDS:
        LengthFmt().print_all(COMMANDS[args[TOPIC]].__doc__)
    else:
        LengthFmt().print_all(commands(COMMANDS))




from urwid import Text, WEIGHT, Columns

from ch2.squeal.tables.statistic import StatisticJournal
from ch2.stoats.calculate.summary import SummaryStatistics
from ch2.uweird.fields import PAGE_WIDTH
from . import Field
from ...lib.utils import label, em


class SummaryField(Field):

    def __init__(self, log, journal, summary=None, width=1):
        if summary is None:
            raise Exception('Summary without summary')
        self._summary = summary
        super().__init__(log, journal, width=width)


class GenericField(SummaryField):

    def __init__(self, log, journal, summary=None, width=1):
        if summary is None:
            raise Exception('Summary without summary')
        self._summary = summary
        super().__init__(log, journal, summary=summary, width=width)

    def _format_value(self, value):
        return self._journal.formatted()

    def _widget(self):
        from urwid import Text
        return Text([label('%s: ' % self._summary), em(self._journal.formatted()), ' '])


def summary_columns(log, s, f, date, schedule, names):

    def fill(columns, width):
        while width < PAGE_WIDTH:
            columns.append(Text(''))
            width += 1

    def field_columns(name):
        journals = StatisticJournal.at_interval(s, date, schedule,
                                                # id of source field is constraint for summary
                                                SummaryStatistics, name.id,
                                                SummaryStatistics)
        columns, width = [], 0
        for named, journal in enumerate(journals):
            summary, period, name = SummaryStatistics.parse_name(journal.statistic_name.name)
            if not named:
                columns.append(Text([name]))
                width += 1
            display = GenericField(log, journal, summary=summary)
            columns.append((WEIGHT, display.width, f(display.widget())))
            width += display.width
        return columns, width

    columns, width = [], 0
    for name in names:
        next_columns, next_width = field_columns(name)
        if width + next_width > PAGE_WIDTH:
            fill(columns, width)
            yield Columns(columns)
            columns, width = [], 0
        columns += next_columns
        width += next_width

    if columns:
        fill(columns, width)
        yield Columns(columns)